@@ -1,7 +1,5 @@
 # wrap VipsImage
 
-from __future__ import division
-
 import numbers
 import struct
 
@@ -83,26 +81,6 @@ def _run_cmplx(fn, image):
     return image
 
 
-# https://stackoverflow.com/a/22409540/1480019
-# https://github.com/benjaminp/six/blob/33b584b2c551548021adb92a028ceaf892deb5be/six.py#L846-L861
-def _with_metaclass(metaclass):
-    """Class decorator for creating a class with a metaclass."""
-    def wrapper(cls):
-        orig_vars = cls.__dict__.copy()
-        slots = orig_vars.get('__slots__')
-        if slots is not None:
-            if isinstance(slots, str):
-                slots = [slots]
-            for slots_var in slots:
-                orig_vars.pop(slots_var)
-        orig_vars.pop('__dict__', None)
-        orig_vars.pop('__weakref__', None)
-        if hasattr(cls, '__qualname__'):
-            orig_vars['__qualname__'] = cls.__qualname__
-        return metaclass(cls.__name__, cls.__bases__, orig_vars)
-    return wrapper
-
-
 # decorator to set docstring
 def _add_doc(name):
     try:
@@ -126,7 +104,7 @@ def _deprecated(note):
     return _dep
 
 
-# Rules for the mapping betweeen numpy typestrings and libvips formats
+# Rules for the mapping between numpy typestrings and libvips formats
 # b1: bool. Above u1 so that rev. map is uchar->u1
 TYPESTR_TO_FORMAT = {'|b1': 'uchar',
                      '|u1': 'uchar',
@@ -203,7 +181,7 @@ def _guess_interpretation(bands, format):
     """
 
     if format not in FORMAT_TO_TYPESTR:
-        raise ValueError('Unknown format: {}'.format(format))
+        raise ValueError(f'Unknown format: {format}')
     if not isinstance(bands, int) or bands < 1:
         raise ValueError('Number of bands must be a positive integer.')
 
@@ -258,8 +236,7 @@ class ImageType(type):
         return call_function
 
 
-@_with_metaclass(ImageType)
-class Image(pyvips.VipsObject):
+class Image(pyvips.VipsObject, metaclass=ImageType):
     """Wrap a VipsImage object.
 
     """
@@ -346,7 +323,7 @@ class Image(pyvips.VipsObject):
 
         pointer = vips_lib.vips_foreign_find_load(vips_filename)
         if pointer == ffi.NULL:
-            raise Error('unable to load from file {0}'.format(vips_filename))
+            raise Error(f'unable to load from file {vips_filename}')
         name = _to_string(pointer)
 
         return pyvips.Operation.call(name, filename,
@@ -516,8 +493,7 @@ class Image(pyvips.VipsObject):
             if ndim > 3:
                 raise ValueError('array has more than 3 dimensions')
             if typestr not in TYPESTR_TO_FORMAT:
-                raise ValueError('conversion from {0} not supported'
-                                 .format(typestr))
+                raise ValueError(f'conversion from {typestr} not supported')
 
             if ndim == 0:
                 width = 1
@@ -610,12 +586,6 @@ class Image(pyvips.VipsObject):
         """
         format_value = GValue.to_enum(GValue.format_type, format)
         pointer = ffi.from_buffer(data)
-        # py3:
-        #   - memoryview has .nbytes for number of bytes in object
-        #   - len() returns number of elements in top array
-        # py2:
-        #   - buffer has no nbytes member
-        #   - but len() gives number of bytes in object
         nbytes = data.nbytes if hasattr(data, 'nbytes') else len(data)
         vi = vips_lib.vips_image_new_from_memory(pointer,
                                                  nbytes,
@@ -664,8 +634,15 @@ class Image(pyvips.VipsObject):
             raise Error('unable to load from source')
         name = _to_string(pointer)
 
-        return pyvips.Operation.call(name, source,
-                                     string_options=options, **kwargs)
+        image = pyvips.Operation.call(name, source,
+                                      string_options=options, **kwargs)
+
+        # keep a secret ref to the source object .. we need that to stay
+        # alive, since it might be a custom source that triggers a python
+        # callback
+        image._references.append(source)
+
+        return image
 
     @staticmethod
     def new_temp_file(format):
@@ -798,7 +775,7 @@ class Image(pyvips.VipsObject):
 
         pointer = vips_lib.vips_foreign_find_save(vips_filename)
         if pointer == ffi.NULL:
-            raise Error('unable to write to file {0}'.format(vips_filename))
+            raise Error(f'unable to write to file {vips_filename}')
         name = _to_string(pointer)
 
         return pyvips.Operation.call(name, self, filename,
@@ -842,6 +819,7 @@ class Image(pyvips.VipsObject):
         format_string = _to_bytes(format_string)
 
         filename = vips_lib.vips_filename_get_filename(format_string)
+        filename = ffi.gc(filename, glib_lib.g_free)
 
         pointer = vips_lib.vips_filename_get_options(format_string)
         options = _to_string_copy(pointer)
@@ -1004,6 +982,29 @@ class Image(pyvips.VipsObject):
 
     # get/set metadata
 
+    def get_gainmap(self, name):
+        """Get the image gainmap, if any.
+
+        Get the gainmap associated with this image, or None for no gainmap.
+        After updating the gainmap, use :meth:`.set_type` to update the
+        `"gainmap"` metadata item.
+
+        Returns:
+            The gainmap image.
+
+        Raises:
+            None
+
+        """
+        if not at_least_libvips(8, 18):
+            return None
+        else:
+            pointer = vips_lib.vips_image_get_gainmap(self.pointer)
+            if pointer == ffi.NULL:
+                return None
+            else:
+                return pyvips.Image(pointer)
+
     def get_typeof(self, name):
         """Get the GType of an item of metadata.
 
@@ -1059,7 +1060,7 @@ class Image(pyvips.VipsObject):
         result = vips_lib.vips_image_get(self.pointer, _to_bytes(name),
                                          gv.pointer)
         if result != 0:
-            raise Error('unable to get {0}'.format(name))
+            raise Error(f'unable to get {name}')
 
         return gv.get()
 
@@ -1128,8 +1129,8 @@ class Image(pyvips.VipsObject):
         """
         gtype = self.get_typeof(name)
         if gtype == 0:
-            raise Error('metadata item {0} does not exist - '
-                        'use set_type() to create and set'.format(name))
+            raise Error(f'metadata item {name} does not exist -'
+                        ' use set_type() to create and set')
         self.set_type(gtype, name, value)
 
     def remove(self, name):
@@ -1164,7 +1165,7 @@ class Image(pyvips.VipsObject):
 
         row_els = self.width if not is_complex else 2 * self.width
 
-        rowfmt = '{0}{1}'.format(row_els, FORMAT_TO_PYFORMAT[self.format])
+        rowfmt = f'{row_els}{FORMAT_TO_PYFORMAT[self.format]}'
         buf = self.write_to_memory()
 
         lst = [list(row) for row in struct.iter_unpack(rowfmt, buf)]
@@ -1204,13 +1205,15 @@ class Image(pyvips.VipsObject):
     #
     #     return interface
 
-    def __array__(self, dtype=None):
+    def __array__(self, dtype=None, copy=None):
         """Conversion to a NumPy array.
 
         Args:
             dtype (str or numpy dtype, optional) The dtype to use for the
                 numpy array. If None, the default dtype of the image is used
                 as defined the global `FORMAT_TO_TYPESTR` dictionary.
+            copy (bool, optional) If True, always copy the data; if None, copy
+                if required; if False, never copy.
 
         Returns:
             numpy.ndarray: The array representation of the image.
@@ -1230,6 +1233,10 @@ class Image(pyvips.VipsObject):
         """
         import numpy as np
 
+        # we always generate a new image, so if copy is False, we can't work
+        if not copy:
+            raise ValueError
+
         arr = (
             np.frombuffer(self.write_to_memory(),
                           dtype=FORMAT_TO_TYPESTR[self.format])
@@ -1282,9 +1289,8 @@ class Image(pyvips.VipsObject):
             array = self.tolist()
             return repr(array)
         else:
-            return ('<pyvips.Image {0}x{1} {2}, {3} bands, {4}>'.
-                    format(self.width, self.height, self.format, self.bands,
-                           self.interpretation))
+            return (f'<pyvips.Image {self.width}x{self.height} {self.format}, '
+                    f'{self.bands} bands, {self.interpretation}>')
 
     def __getattr__(self, name):
         """Divert unknown names to libvips.
@@ -1430,8 +1436,8 @@ class Image(pyvips.VipsObject):
 
             # n.b. We don't use isinstance because isinstance(True, int)
             # is True!
-            if not (all(type(x) == int for x in arg) or
-                    all(type(x) == bool for x in arg)):
+            if not (all(type(x) == int for x in arg) or  # noqa: E721
+                    all(type(x) == bool for x in arg)):  # noqa: E721
                 raise IndexError('list must contain only ints or only bools')
 
             if isinstance(arg[0], bool):
@@ -1879,6 +1885,8 @@ class Image(pyvips.VipsObject):
         """Add an alpha channel."""
         if self.interpretation == 'grey16' or self.interpretation == 'rgb16':
             max_alpha = 65535
+        elif self.interpretation == 'scrgb':
+            max_alpha = 1.0
         else:
             max_alpha = 255
         return self.bandjoin(max_alpha)
